//EEE3050 Theory on Computer Architectures (Spring 2017)

//Instructor: Prof. Jinkyu Jeong (jinkyu@skku.edu)

//TA: Gyusun Lee(lgs0409@naver.com), Minwoo Ahn(mwahn402@gmail.com)

//Semiconductor Building #400509

//Author: Minwoo Ahn

//Copyright (c) 2017 SungKyunKwan Univ. CSL(Computer Systems Laboratory)

#include "mem.h"

/* Global Variables */
int hit_count = 0;
int miss_count = 0;

/* Variables for hw3_check */
int write_count = 0;
int wb_count = 0;
unsigned int write_addr[100000];
char write_value[100000][4];

int main(){
	char input_file[20];
	char output_file[50];
	FILE *fd_input;
	FILE *fd_output;
	char op;		//Read or Write of request
	unsigned int phy_addr;
	char trash;
	char write[4];
	int i, j;

	/* Get Input */
	printf("====================Welcome to Cache Simulation!=====================\n");
	printf("Enter # of Ways : ");
	scanf("%d", &cache.num_ways);
	printf("Enter # of Entries : ");
	scanf("%d", &cache.num_entry);
	printf("Enter # of blocks per entry : ");
	scanf("%d", &cache.block_per_entry);
	
	cache.bytes_per_block = 4;
	
	initialize();

	printf("Input file name : ");
	scanf("%s", input_file);


	/* Open Input & Output Files */
	fd_input = fopen(input_file, "r");
	if(fd_input == NULL){
		printf("There is no file such %s\n", input_file);
		exit(1);
	}

	strcpy(output_file, "cache_trace_");
	strcat(output_file, input_file);

	fd_output = fopen(output_file, "w+");
	

	/* Start Simulation */
	while(1){
		fscanf(fd_input, "%c", &op);
		
		switch(op){
			case 'R':	//Read Operation
				fscanf(fd_input, "%u", &phy_addr);
				fprintf(fd_output, "READ, %u", phy_addr);
				fprintf(fd_output, ", value : %s\n", read_op(phy_addr));
				break;
			case 'W':	//Write Operation
				fscanf(fd_input, "%u", &phy_addr);
				fscanf(fd_input, "%s", write);
				write_addr[write_count] = phy_addr;
				memcpy(write_value[write_count], write, 4);
				fprintf(fd_output, "WRITE, %u, %s\n", phy_addr, write);
				write_count++;
				write_op(phy_addr, write);
				break;
			default:	//HALT
				goto end;
		};
		print_cache(fd_output);
		fscanf(fd_input, "%c", &trash);
	}

end:
	fclose(fd_input);

	printf("\n\n");
	for(i = 0; i < cache.num_entry; i++){
		for(j = 0; j < cache.num_ways; j++){
			if(cache.way_cache[j][i].dirty == 1){
				flush(j, i);
				wb_count++;
			}
		}
	}
	if(!hw3_check())	printf("[Error!!] Write values are different with data in memory.\n");
	else	printf("[Correct!!]\n");

	fprintf(fd_output, "==============================Final Result=============================\n");
	print_cache(fd_output);


	/*	Print Results	  */
	fprintf(fd_output, "Total Write Back by Dirty Bit Count:	%d\n\n", wb_count);
	fprintf(fd_output, "Total Hit Count :	%d\n", hit_count);
	fprintf(fd_output, "Total Miss Count :	%d\n\n", miss_count);

	fprintf(fd_output, "Hit Ratio : %.3f\n", (double)hit_count / (double)(hit_count + miss_count));

	destroy();

	return 0;
}

void initialize(){
	int i, j, k, l;
	
	printf("Initialize Your Cache...\n");
	/* cache[num_ways][num_entry]*/
	cache.way_cache = (struct cache_entry **)malloc(sizeof(struct cache_entry *) * cache.num_ways);
	
	for(i = 0; i < cache.num_ways; i++)	cache.way_cache[i] = (struct cache_entry *)malloc(sizeof(struct cache_entry) * cache.num_entry);
	
	for(i = 0; i < cache.num_ways; i++){
		for(j = 0; j < cache.num_entry; j++){
			cache.way_cache[i][j].data = (char **)malloc(sizeof(char *) * cache.block_per_entry);
			
			for(k = 0; k < cache.block_per_entry; k++){
				cache.way_cache[i][j].data[k] = (char *)malloc(sizeof(char) * cache.bytes_per_block);
				memset(cache.way_cache[i][j].data[k], 0, cache.bytes_per_block);
			}
			/* Initialize other variables in your cache_entry structure! */
			/* e.g. valid, dirty, tag, etc.                              */
			cache.way_cache[i][j].valid = 0;
			cache.way_cache[i][j].tag = -1;
			cache.way_cache[i][j].dirty = 0;
			///////////////////////////////////////////////////////////////
		}
	}

	/* Disk Data Initialization */
	for(i = 0; i < 1024 * 1024 * 256; i++){
		memory[i][0] = i % 94 + 33;
		memory[i][1] = (i + 1) % 94 + 33;
		memory[i][2] = (i + 2) % 94 + 33;
		memory[i][3] = (i + 3) % 94 + 33;
	}
	
	printf("Finished to Initialize Your Cache!\n");
}

void destroy(){
	int i, j, k, l;

	printf("Destroy Your Cache...\n");

	for(i = 0; i < cache.num_ways; i++){
		for(j = 0; j < cache.num_entry; j++){
			for(k = 0; k < cache.block_per_entry; k++)	free(cache.way_cache[i][j].data[k]);
			free(cache.way_cache[i][j].data);
		}
	}

	for(i = 0; i < cache.num_ways; i++)	free(cache.way_cache[i]);
	free(cache.way_cache);

	printf("Finished to Destroy Your Cache!\n");
}

int hw3_check(){
	char final_value[100000][4];
	int finished[100000];
	int i;
	int j;
	int index;
	int finish_count = 0;
	int check;

	for(i = 0; i < write_count; i++){
		check = 0;

		for(j = 0; j < finish_count; j++){
			if(finished[j] == write_addr[i]){
				check = 1;
				break;
			}
		}

		if(check == 0){
			index = i;
			for(j = i + 1; j < write_count; j++){
				if(write_addr[i] == write_addr[j])	index = j;	
			}

			if(memcmp(memory[write_addr[index] / 4], write_value[index], 4) != 0)	return 0;
			
			finished[finish_count] = write_addr[j];
			finish_count++;
		}
	}	

	return 1;	
}

char *read_op(int addr){
	/* You have to implement your own read_op function(2-way set associative) here! */

	
	//////////////////////////////////////////////////////////////////////////////////
	return "NONE";
}

void write_op(int addr, char *write){
	/* You have to implement your own write_op function(2-way set associative) here! */


	///////////////////////////////////////////////////////////////////////////////////
}

/* Write data in cache(way, index) to disk */
void flush(int way, int index){
	/* You have to implement your own flush function using dirty bits here! */


	//////////////////////////////////////////////////////////////////////////
}

void print_cache(FILE *in){
	int i, j, k, l;
	int bytes_per_block;
	int block_per_entry;
	struct cache_entry *c;

	bytes_per_block = cache.bytes_per_block;
	block_per_entry = cache.block_per_entry;

	for(i = 0; i < cache.num_ways; i++){
		fprintf(in, "[Way %d]\n", i);
		fprintf(in, "Valid\tDirty\tTag\t");
		for(j = 0; j < block_per_entry; j++){
			fprintf(in, "Data[%d]\t", j);
		}
		fprintf(in, "\n");
		for(j = 0; j < cache.num_entry; j++){
			c = &cache.way_cache[i][j];
			fprintf(in, "%d\t%d\t%d", c -> valid, c -> dirty, c -> tag);
		
			for(k = 0; k < block_per_entry; k++){
				fprintf(in, "\t");
				for(l = 0; l < bytes_per_block; l++){
					if(c -> data[k][l] != 0)	fprintf(in, "%c", c -> data[k][l]);
					else	fprintf(in, " ");
				}
			}
			fprintf(in, "\n");
		}
		fprintf(in, "\n");
	}
	fprintf(in, "\n\n\n");
}
